/* David Leonard, 2003. Public domain. */
/* $Id: xsolitaire.c,v 1.10 2003/05/02 06:23:26 d Exp $ */

#include <stdio.h>
#include <stdlib.h>
#include <sys/time.h>
#include <X11/IntrinsicP.h>
#include <X11/CompositeP.h>
#include <X11/Intrinsic.h>
#include <X11/StringDefs.h>
#include <X11/Shell.h>
#include <X11/Xaw/Box.h>
#include <X11/Xaw/Command.h>
#include <X11/Xaw/Label.h>
#include <X11/Xaw/Paned.h>
#include <X11/Xaw/Tip.h>
#include <X11/Xmu/Editres.h>
#include <X11/Xmu/SysUtil.h>

#include "Stack.h"

#define	NPLAY 	7		/* Number of play stacks */
#define NORDER	4		/* Number of solve stacks */

/*
 * Red 3 follows Black 2, etc. in alternating colour
 */
#define FOLLOWS(a,b)	((a) && (b) && CARD_COLOUR(a) != CARD_COLOUR(b) && \
			 CARD_FACE(a) == CARD_FACE(b) + 1)

static Boolean score_draw_three = True;
static Boolean score_cumulative = False;
static enum { SCORE_STANDARD, SCORE_VEGAS } score_system = SCORE_STANDARD;
static Cardinal deck_exhausts;

enum undo { UNDO_NONE, UNDO_HIT, UNDO_HITNEW, UNDO_DRAG, UNDO_FLIP };

static void Quit(Widget, XEvent *, String *, Cardinal *);
static void Deal(Widget, XEvent *, String *, Cardinal *);
static void Undo(Widget, XEvent *, String *, Cardinal *);
static void AutoMove(Widget, XEvent *, String *, Cardinal *);
static void Win(Widget, XEvent *, String *, Cardinal *);

static void QuitCallback(Widget, XtPointer, XtPointer);
static void DealCallback(Widget, XtPointer, XtPointer);
static void UndoCallback(Widget, XtPointer, XtPointer);

static void HitDrawPile(Widget, XtPointer, XtPointer);
static Boolean SplitDiscard(Widget, XtPointer, Cardinal);
static Boolean SplitOrdered(Widget, XtPointer, Cardinal);
static Boolean DropOrdered(Widget, XtPointer, CardID *, Cardinal, Widget);
static Boolean SplitPlay(Widget, XtPointer, Cardinal);
static Boolean DropPlay(Widget, XtPointer, CardID *, Cardinal, Widget);
static void HitPlay(Widget, XtPointer, XtPointer);
static void DropNotify(Widget, XtPointer, XtPointer);
static void SetUndoType(enum undo);

static void RewardStart(void);
static void RewardStop(void);
static void RewardProc(XtPointer, XtIntervalId *);
static void UpdateTime(XtPointer, XtIntervalId *);
static void SetTime(unsigned int);
static void AddScore(int);

static String fallback_resources[] = {
	NULL
};

static XtActionsRec actions[] = {
	{ "quit",	  Quit },
	{ "deal",	  Deal },
	{ "undo",	  Undo },
	{ "win",	  Win },
	{ "auto-move",	  AutoMove },
};

static XtAppContext app_context;
static Widget toplevel;
static Widget drawpile, discard, ordered[NORDER], play[NPLAY];
static Widget message, undo_command;
static Widget score_widget, time_widget;
static Boolean game_won;
static XtIntervalId reward_timer, elapsed_timer;
static time_t start_time;
static int score;

/*
 * Undo tracking
 *	none	- nothing to undo
 *	hit	- replace 'count' cards onto drawpile
 *	hitnew	- move drawpile back to discard pile
 *	drag	- move 'count' cards from 'target' back to 'source'
 *	flip	- flip top card of play stack 'source' back down
 */
static enum undo undo_type;
static Widget undo_source, undo_target;
static Cardinal undo_count;
static int undo_score;

/*
 * Main program
 */
int
main(argc, argv)
	int argc;
	char *argv[];
{
	Widget pane, cmdbox, solvebox, playbox, scorebox, popup, cmd, ignore; 
	Arg args[10];
	int nargs;
	int i;

	toplevel = XtOpenApplication(
	    &app_context, 		/* app_context_return */
	    "Solitaire", 		/* application_class */
	    NULL, 0,			/* options, num_options */
	    &argc, argv,		/* argc_in_out, argv_in_out */
	    fallback_resources,		/* fallback_resources */
	    sessionShellWidgetClass,	/* widget_class */
	    NULL, 0);			/* args, num_args */

	XtAppAddActions(app_context, actions, XtNumber(actions));

	/*
	 * Create the widgets
	 */
	pane = XtCreateManagedWidget("pane", panedWidgetClass, toplevel,
		NULL,0);

	cmdbox = XtCreateManagedWidget("cmdbox", boxWidgetClass, pane,
		NULL,0);
	solvebox = XtCreateManagedWidget("solvebox", boxWidgetClass, pane, 
		NULL, 0);
	playbox = XtCreateManagedWidget("playbox", boxWidgetClass, pane,
		NULL, 0);
	scorebox = XtCreateManagedWidget("scorebox", boxWidgetClass, pane,
		NULL, 0);

	/* The draw pile */
	nargs = 0;
	XtSetArg(args[nargs], XtNsplitCallback, StackNeverSplit); nargs++;
	XtSetArg(args[nargs], XtNdropCallback, StackNeverDrop); nargs++;
	drawpile = XtCreateManagedWidget("drawpile", stackWidgetClass,
		solvebox, args, nargs);
	XtAddCallback(drawpile, XtNhitNotify, HitDrawPile, NULL);

	/* The discard pile */
	nargs = 0;
	XtSetArg(args[nargs], XtNdropCallback, StackNeverDrop); nargs++;
	XtSetArg(args[nargs], XtNsplitCallback, SplitDiscard); nargs++;
	discard = XtCreateManagedWidget("discard", stackWidgetClass, solvebox, 
		args, nargs);

	/* A space/placeholder*/
	nargs = 0;
	XtSetArg(args[nargs], XtNdropCallback, StackNeverDrop); nargs++;
	XtSetArg(args[nargs], XtNsplitCallback, StackNeverSplit); nargs++;
	ignore = XtCreateManagedWidget("ignore", stackWidgetClass, solvebox, 
		args, nargs);

	/* The four ordered/solved stacks */
	nargs = 0;
	XtSetArg(args[nargs], XtNdropCallback, DropOrdered); nargs++;
	XtSetArg(args[nargs], XtNsplitCallback, SplitOrdered); nargs++;
	for (i = 0; i < NORDER; i++) {
		ordered[i] = XtCreateManagedWidget("ordered",
				stackWidgetClass, solvebox, args, nargs);
		XtAddCallback(ordered[i], XtNdropNotify, DropNotify, NULL);
	}

	/* The seven play stacks */
	nargs = 0;
	XtSetArg(args[nargs], XtNdropCallback, DropPlay); nargs++;
	XtSetArg(args[nargs], XtNsplitCallback, SplitPlay); nargs++;
	for (i = 0; i < NPLAY; i++) {
		play[i] = XtCreateManagedWidget("play",
				stackWidgetClass, playbox, args, nargs);
		XtAddCallback(play[i], XtNhitNotify, HitPlay, NULL);
		XtAddCallback(play[i], XtNdropNotify, DropNotify, NULL);
	}

	/* The top row of messages and buttons */

	message = XtCreateManagedWidget("message", labelWidgetClass, cmdbox, 
	    NULL, 0);
	XawTipEnable(message);

	cmd = XtCreateManagedWidget("quit", commandWidgetClass, cmdbox, 
	    NULL, 0);
	XtAddCallback(cmd, XtNcallback, QuitCallback, NULL);
	XawTipEnable(cmd);

	cmd = XtCreateManagedWidget("deal", commandWidgetClass, cmdbox, 
	    NULL, 0);
	XtAddCallback(cmd, XtNcallback, DealCallback, NULL);
	XawTipEnable(cmd);

	cmd = XtCreateManagedWidget("undo", commandWidgetClass, cmdbox, 
	    NULL, 0);
	XtAddCallback(cmd, XtNcallback, UndoCallback, NULL);
	XawTipEnable(cmd);
	undo_command = cmd;

	/* The bottom row of scores and timers */
	XtCreateManagedWidget("timeLabel", labelWidgetClass, scorebox, 
	    NULL, 0);
	time_widget = XtCreateManagedWidget("time", labelWidgetClass, scorebox, 
	    NULL, 0);
	XtCreateManagedWidget("scoreLabel", labelWidgetClass, scorebox, 
	    NULL, 0);
	score_widget = XtCreateManagedWidget("score", labelWidgetClass,
	    scorebox, NULL, 0);

	/* Initialise drag-and-drop for stacks */
	popup = StackCreateDragAndDropPopup(toplevel, "popup");

	/* Deal a new game */
	i = 0;
	Deal(toplevel, NULL, NULL, &i);

	XtAppAddTimeOut(app_context, 1000, UpdateTime, NULL);

	XtRealizeWidget(toplevel);
	XtAddEventHandler(toplevel, 0, True, _XEditResCheckMessages, NULL);

	/* Main application loop */
	XtAppMainLoop(app_context);
	exit(0);
}

/*------------------------------------------------------------
 * Actions
 */

/*
 * Quit the game
 */
static void
Quit(w, event, args, arg_count)
	Widget w;
	XEvent *event;
	String *args;
	Cardinal *arg_count;
{
	XtDestroyWidget(toplevel);
	XtAppSetExitFlag(app_context);
}

/*
 * Reset game and deal a new board
 */
static void
Deal(w, event, args, arg_count)
	Widget w;
	XEvent *event;
	String *args;
	Cardinal *arg_count;
{
	int i, j, c;
	CardID cards[CARD_STANDARD_DECK_SIZE];

	/* Stop any winning reward thing */
	if (game_won) {
		game_won = False;
		RewardStop();
	}

	/* Kill the elapsed timer */
	if (elapsed_timer) {
		XtRemoveTimeOut(elapsed_timer);
		elapsed_timer = 0;
	}
	SetTime(0);
	start_time = 0;

	/* Forget any undo history */
	SetUndoType(UNDO_NONE);

	/* Reset the table */
	StackClear(drawpile);
	StackClear(discard);
	for (i = 0; i < NORDER; i++)
		StackClear(ordered[i]);
	for (i = 0; i < NPLAY; i++)
		StackClear(play[i]);

	/* Select the standard 52 cards and shuffle them */
	CardFillStandardDeck(cards);
	CardShuffle(cards, XtNumber(cards));

	/* Lay out the play cards */
	c = 0;
	for (i = 0; i < NPLAY; i++)
	    for (j = i; j < NPLAY; j++)
		StackPush(play[j], cards[c++] | (i != j ? CARD_DOWN : 0));

	/* Put the remaining cards on the drawpile */
	while (c < XtNumber(cards))
		StackPush(drawpile, cards[c++] | CARD_DOWN);

	/* Reset the score */
	if (!score_cumulative) {
		score = 0;
		AddScore(0);	/* Force display update */
	}
}

/*
 * Undo the last action, if it was recorded
 */
static void
Undo(Widget w, XEvent *event, String *params, Cardinal *num_params)
{
	int i;
	CardID card;

	switch (undo_type) {
	default:
	case UNDO_NONE:
		XBell(XtDisplay(w), 0);
		break;
	case UNDO_HIT:
		for (i = 0; i < undo_count; i++) {
			card = StackPop(discard);
			StackPush(drawpile, card | CARD_DOWN);
		}
		break;
	case UNDO_HITNEW:
		while ((card = StackPop(drawpile)) != CARD_NONE)
			StackPush(discard, card & ~CARD_DOWN);
		break;
	case UNDO_FLIP:
		card = StackPop(undo_source);
		StackPush(undo_source, card | CARD_DOWN);
		break;
	case UNDO_DRAG:
		StackMove(undo_source, undo_target, undo_count);
		break;
	}
	if (undo_type != UNDO_NONE)
		score = undo_score;
	SetUndoType(UNDO_NONE);	/* don't allow redo */
}

/*
 * Automatically move as many cards from the play stacks onto 
 * the solved stacks as we can
 */
static void
AutoMove(w, event, args, arg_count)
	Widget w;
	XEvent *event;
	String *args;
	Cardinal *arg_count;
{
	Boolean moved_something = False;
	int i, j;
	CardID card, top;

    again:
	if (game_won) 
		return;
	for (i = 0; i < NPLAY; i++) {
	    card = StackTop(play[i]);
	    if (card != CARD_NONE) {
		if (CARD_IS_DOWN(card)) {
		    HitPlay(play[i], NULL, NULL);
		    moved_something = True;
		    card = StackTop(play[i]);
		}
		for (j = 0; j < NORDER; j++) {
			top = StackTop(ordered[j]);
			if ((top != CARD_NONE && 
			    CARD_SUIT(top) == CARD_SUIT(card) &&
			    CARD_FACE(top) + 1 == CARD_FACE(card)) ||
			    (top == CARD_NONE && CARD_FACE(card) == 1))
			{
				StackDropEvent ev;
				StackMove(ordered[j], play[i], 1);
				ev.source = play[i];
				ev.target = ordered[j];
				ev.count = 1;
				DropNotify(ordered[j], NULL, &ev);
				moved_something = True;
				SetUndoType(UNDO_NONE);
				goto again;
			}
		}
	    }
	}

	if (!moved_something)
		XBell(XtDisplay(w), 0);
}

/*
 * Win procedure
 */
static void
Win(w, event, args, arg_count)
	Widget w;
	XEvent *event;
	String *args;
	Cardinal *arg_count;
{
	if (!game_won) {
		SetUndoType(UNDO_NONE);		/* unreversible */
		game_won = True;

		XBell(XtDisplay(w), 0);		/* pathetic */
		RewardStart();			/* almost as pathetic */
	}
}

/*------------------------------------------------------------
 * Command callbacks
 */

static void
QuitCallback(widget, closure, call_data)
	Widget widget;
	XtPointer closure, call_data;
{
	Quit(widget, NULL, NULL, NULL);
}

static void
DealCallback(widget, closure, call_data)
	Widget widget;
	XtPointer closure, call_data;
{
	Deal(widget, NULL, NULL, NULL);
}

static void
UndoCallback(widget, closure, call_data)
	Widget widget;
	XtPointer closure, call_data;
{
	Undo(widget, NULL, NULL, NULL);
}

/*------------------------------------------------------------
 * Stack callbacks
 */

/*
 * User clicked on the draw pile for some new cards
 */
static void
HitDrawPile(Widget w, XtPointer closure, XtPointer data)
{
	CardID card;
	Boolean changed = False;

	if (StackDepth(drawpile) == 0) {
		while ((card = StackPop(discard)) != CARD_NONE) {
			StackPush(drawpile, card | CARD_DOWN);
			changed = True;
		}
		if (changed) {
			SetUndoType(UNDO_HITNEW);
			deck_exhausts++;
			if (score_system == SCORE_STANDARD) {
			    if (score_draw_three) {
				if (deck_exhausts > 3)
				    AddScore(-20);
			    } else {
				if (deck_exhausts > 1)
				    AddScore(-100);
			    }
			}
		}
	} else {
		int count = 0;
		while (count < 3 && (card = StackPop(drawpile)) != CARD_NONE) {
			StackPush(discard, card & ~CARD_DOWN);
			count++;
			changed = True;
		}
		if (changed) {
			undo_count = count;
			SetUndoType(UNDO_HIT);
		}
	}
}

/*
 * User wants to pick up some cards from the discard stack
 */
static Boolean
SplitDiscard(w, d, depth)
	Widget w;
	XtPointer d;
	Cardinal depth;
{
	return depth == 1;
}

/*
 * User wants to pick up some cards from one of the four solved stacks
 */
static Boolean
SplitOrdered(w, d, depth)
	Widget w;
	XtPointer d;
	Cardinal depth;
{
	if (game_won)
		return False;
	return depth == 1;
}

/*
 * User wants to drop some cards onto one of the four solved stack
 */
static Boolean
DropOrdered(w, d, cards, num_cards, source)
	Widget w;
	XtPointer d;
	CardID *cards;
	Cardinal num_cards;
	Widget source;
{
	CardID top, card;

	if (num_cards != 1) 
		return False;
	top = StackPeek(w, 0);
	card = cards[0];
	if (top == CARD_NONE && CARD_FACE(card) == FACE_ACE)
		return True;
/*printf("DropOrdered: card=0x%x(%c%c) top=0x%x(%c%c)\n", card, *CardFaceName(card), *CardSuitName(card), top, *CardFaceName(top), *CardSuitName(top));*/
	if (CARD_SUIT(card) == CARD_SUIT(top) &&
	    CARD_FACE(card) == CARD_FACE(top) + 1)
		return True;
	return False;
}

/*
 * User wants to pick up cards from a play stack
 */
static Boolean
SplitPlay(w, d, depth)
	Widget w;
	XtPointer d;
	Cardinal depth;
{
	CardID card, prev;
	int pos;

	if (game_won)
		return False;

	/* Don't allow any splits if the top card is down */
	card = StackPeek(w, 0);
	if (CARD_IS_DOWN(card))
		return False;

	/* Always allow top card to be picked */
	if (depth == 1)
		return True;
	prev = card;
	pos = 1;
	while (pos < depth) {
		card = StackPeek(w, pos);
		if (CARD_IS_DOWN(card))
			return False;
		if (!FOLLOWS(card, prev))
			return False;
		prev = card;
		pos++;
	}
	return True;
}

/*
 * User wants to drop some cards on a play stack
 */
static Boolean
DropPlay(w, d, cards, num_cards, source)
	Widget w;
	XtPointer d;
	CardID *cards;
	Cardinal num_cards;
	Widget source;
{
	CardID bottom, top;

	bottom = cards[num_cards-1];
	top = StackPeek(w, 0);

/*printf("DropPlay: bottom=0x%x(%c%c) top=0x%x(%c%c)\n", bottom, *CardFaceName(bottom), *CardSuitName(bottom), top, *CardFaceName(top), *CardSuitName(top));*/

	if (top == CARD_NONE && CARD_FACE(bottom) == FACE_KING)
		return True;
	if (FOLLOWS(top, bottom))
		return True;
	return False;
}

/*
 * User hit a play card (that was probably face down)
 */
static void
HitPlay(Widget w, XtPointer closure, XtPointer data)
{
	CardID card;

	/* Flip the top card if it is face down */
	card = StackPeek(w, 0);
	if (CARD_IS_DOWN(card)) {
		card = StackPop(w);
		StackPush(w, card & ~CARD_DOWN);
		undo_source = w;
		SetUndoType(UNDO_FLIP);
		if (score_system == SCORE_STANDARD) {
		    /* If you turn over a card in a row stack, 
		     * you get 5 points. */
		    AddScore(+5);
		}
	}
}

/*
 * Remember drops so we can undo them
 */
static void
DropNotify(Widget w, XtPointer closure, XtPointer data)
{
	StackDropEvent *ev = (StackDropEvent *)data;
	int i;
	enum { sORDERED, sPLAY, sDISCARD } src, tgt;

	undo_source = ev->source;
	undo_target = ev->target;
	undo_count = ev->count;
	SetUndoType(UNDO_DRAG);

	if (ev->source == discard)				src = sDISCARD;
	else if (strcmp(XtName(ev->source), "play") == 0)	src = sPLAY;
	else if (strcmp(XtName(ev->source), "ordered") == 0)	src = sORDERED;

	if (ev->target == discard)				tgt = sDISCARD;
	else if (strcmp(XtName(ev->target), "play") == 0)	tgt = sPLAY;
	else if (strcmp(XtName(ev->target), "ordered") == 0)	tgt = sORDERED;

	/* Scoring */
	if (score_system == SCORE_STANDARD) {
	    /* If you move a card to a suit stack, you get 10 points */
	    if (tgt == sORDERED)
		    AddScore(+10);
	    /* If you move a card from the deck to a row stack, you 
	     * get 5 points. */
	    if (src == sDISCARD && tgt == sPLAY)
		    AddScore(+5);
	    /* If you move a card from a suit stack back to a row stack, you 
	     * lose 15 points. */
	    if (src == sORDERED && tgt == sPLAY)
		    AddScore(-15);
	}

	/* Check to see if we have won */
	for (i = 0; i < NORDER; i++)
		if (CARD_FACE(StackTop(ordered[i])) != FACE_KING)
			return;
	i = 0;
	Win(w, NULL, NULL, &i);
}

/*
 * Set the current undo type, and update the Undo command widget
 * (Make sure this is called *before* changing the score
 */
static void
SetUndoType(type)
	enum undo type;
{
	undo_type = type;
	undo_score = score;

	if (!start_time && type != UNDO_NONE) {
		/* Kick the elapsed timer off */
		start_time = time(NULL);
		if (!elapsed_timer) 
			UpdateTime(NULL, NULL);
	}

	XtSetSensitive(undo_command, undo_type != UNDO_NONE);
}

/*------------------------------------------------------------
 * Win reward
 * - kind of pathetic, but we simply move the king stacks into
 *   the play area, and then repeatedly highlight/unhighlight them
 */

static int reward_index;

static void
RewardStart()
{
	int i;

	/* Don't do anything if already running */
	if (reward_timer)
		return;

	/* Move the winning cards into the play area */
	for (i = 0; i < NPLAY; i++)
		StackClear(play[i]);
	StackClear(drawpile);
	StackClear(discard);
	StackPush(play[0], CARD_BLACK_JOKER);
	StackPush(play[5], CARD_RED_JOKER);

	reward_index = 0;
	RewardProc(NULL, NULL);		/* Kick the timout proc going */
}

static void
RewardStop()
{
	XtRemoveTimeOut(reward_timer);
	XtCallActionProc(play[reward_index], "unhighlight", NULL, 
		NULL, 0);
	reward_timer = 0;
}

static void
RewardProc(closure, id)
	XtPointer closure;
	XtIntervalId *id;
{
	int i;
	CardID card;

	/* Gradually move cards from the ordered stacks onto play stacks */
	for (i = 0; i < NORDER; i++)
		if (StackDepth(ordered[i])) {
			card = StackPop(ordered[i]);
			StackPush(play[i+1], card);
			break;
		}

	XtCallActionProc(play[reward_index], "unhighlight", NULL, 
		NULL, 0);
	reward_index = (reward_index+1) % NPLAY;
	XtCallActionProc(play[reward_index], "highlight", NULL, 
		NULL, 0);
	reward_timer = XtAppAddTimeOut(app_context, 250, RewardProc, NULL);
}

static void
UpdateTime(closure, id)
	XtPointer closure;
	XtIntervalId *id;
{
	if (start_time)
		SetTime(time(NULL) - start_time);
	if (game_won)
		elapsed_timer = 0;
	else
		elapsed_timer = XtAppAddTimeOut(app_context, 1000, UpdateTime,
			NULL);
}

/* Set the time widget to read the number of elapsed seconds */
static void
SetTime(elapsed)
	unsigned int elapsed;
{
	char buf[] = "0000+00:00";
	unsigned int secs, mins, hrs, days;
	Arg args[1];
	int nargs;

	secs = elapsed % 60; elapsed /= 60;
	mins = elapsed % 60; elapsed /= 60;
	hrs = elapsed % 24; elapsed /= 24;
	days = elapsed;

	if (days)
		XmuSnprintf(buf, sizeof buf, "%d+%02d:%02d:%02d",
			days, hrs, mins, secs);
	else if (hrs)
		XmuSnprintf(buf, sizeof buf, "%2d:%02d:%02d",
			hrs, mins, secs);
	else 
		XmuSnprintf(buf, sizeof buf, "%2d:%02d",
			mins, secs);

	nargs = 0;
	XtSetArg(args[nargs], XtNlabel, buf); nargs++;
	XtSetValues(time_widget, args, nargs);
}

/* Update the score */
static void
AddScore(increment)
	int increment;
{
	char buf[64];
	Arg args[1];
	int nargs;

	score += increment;
	XmuSnprintf(buf, sizeof buf, "%d", score);

	nargs = 0;
	XtSetArg(args[nargs], XtNlabel, buf); nargs++;
	XtSetValues(score_widget, args, nargs);
}
